@@ -1,6 +1,7 @@
 import copy
 import math
 import os
+import re
 import sys
 import string
 import warnings
@@ -12,10 +13,117 @@ import sklearn
 from IPython.core.display import display, HTML
 from sklearn.preprocessing import StandardScaler
 from sklearn.utils import check_random_state, all_estimators
-from sklearn.ensemble import RandomForestClassifier
 from .rulefit import RuleFit
 from statsmodels.stats.outliers_influence import variance_inflation_factor
 from statsmodels.tools.tools import add_constant
+import pickle
+
+
+def AutoSpearman(X_train, correlation_threshold=0.7, correlation_method='spearman', VIF_threshold=5):
+    """An automated feature selection approach that address collinearity and multicollinearity.
+    For more information, please kindly refer to the `paper <https://ieeexplore.ieee.org/document/8530020>`_.
+
+    Parameters
+    ----------
+    X_train : :obj:``
+        Threshold value of correalation.
+    correlation_threshold : :obj:`float`
+        Threshold value of correalation.
+    correlation_method : :obj:`str`
+        Method for solving the correlation between the features.
+    VIF_threshold : :obj:`int`
+        Threshold value of VIF score.
+    """
+    X_AS_train = X_train.copy()
+    AS_metrics = X_AS_train.columns
+    count = 1
+
+    # (Part 1) Automatically select non-correlated metrics based on a Spearman rank correlation test.
+    print('(Part 1) Automatically select non-correlated metrics based on a Spearman rank correlation test')
+    while True:
+        corrmat = X_AS_train.corr(method=correlation_method)
+        top_corr_features = corrmat.index
+        abs_corrmat = abs(corrmat)
+
+        # identify correlated metrics with the correlation threshold of the threshold
+        highly_correlated_metrics = ((corrmat > correlation_threshold) | (corrmat < -correlation_threshold)) & (
+                corrmat != 1)
+        n_correlated_metrics = np.sum(np.sum(highly_correlated_metrics))
+        if n_correlated_metrics > 0:
+            # find the strongest pair-wise correlation
+            find_top_corr = pd.melt(abs_corrmat, ignore_index=False)
+            find_top_corr.reset_index(inplace=True)
+            find_top_corr = find_top_corr[find_top_corr['value'] != 1]
+            top_corr_index = find_top_corr['value'].idxmax()
+            top_corr_i = find_top_corr.loc[top_corr_index, :]
+
+            # get the 2 correlated metrics with the strongest correlation
+            correlated_metric_1 = top_corr_i[0]
+            correlated_metric_2 = top_corr_i[1]
+            print('> Step', count, 'comparing between', correlated_metric_1, 'and', correlated_metric_2)
+
+            # compute their correlation with other metrics outside of the pair
+            correlation_with_other_metrics_1 = np.mean(abs_corrmat[correlated_metric_1][
+                                                           [i for i in top_corr_features if
+                                                            i not in [correlated_metric_1, correlated_metric_2]]])
+            correlation_with_other_metrics_2 = np.mean(abs_corrmat[correlated_metric_2][
+                                                           [i for i in top_corr_features if
+                                                            i not in [correlated_metric_1, correlated_metric_2]]])
+            print('>>', correlated_metric_1, 'has the average correlation of',
+                  np.round(correlation_with_other_metrics_1, 3), 'with other metrics')
+            print('>>', correlated_metric_2, 'has the average correlation of',
+                  np.round(correlation_with_other_metrics_2, 3), 'with other metrics')
+            # select the metric that shares the least correlation outside of the pair and exclude the other
+            if correlation_with_other_metrics_1 < correlation_with_other_metrics_2:
+                exclude_metric = correlated_metric_2
+            else:
+                exclude_metric = correlated_metric_1
+            print('>>', 'Exclude', exclude_metric)
+            count = count + 1
+            AS_metrics = list(set(AS_metrics) - set([exclude_metric]))
+            X_AS_train = X_AS_train[AS_metrics]
+        else:
+            break
+
+    print('According to Part 1 of AutoSpearman,', AS_metrics, 'are selected.')
+
+    # (Part 2) Automatically select non-correlated metrics based on a Variance Inflation Factor analysis.
+    print('(Part 2) Automatically select non-correlated metrics based on a Variance Inflation Factor analysis')
+
+    # Prepare a dataframe for VIF
+    X_AS_train = add_constant(X_AS_train)
+
+    selected_features = X_AS_train.columns
+    count = 1
+    while True:
+        # Calculate VIF scores
+        vif_scores = pd.DataFrame([variance_inflation_factor(X_AS_train.values, i)
+                                   for i in range(X_AS_train.shape[1])],
+                                  index=X_AS_train.columns)
+        # Prepare a final dataframe of VIF scores
+        vif_scores.reset_index(inplace=True)
+        vif_scores.columns = ['Feature', 'VIFscore']
+        vif_scores = vif_scores.loc[vif_scores['Feature'] != 'const', :]
+        vif_scores.sort_values(by=['VIFscore'], ascending=False, inplace=True)
+
+        # Find features that have their VIF scores of above the threshold
+        filtered_vif_scores = vif_scores[vif_scores['VIFscore'] >= VIF_threshold]
+
+        # Terminate when there is no features with the VIF scores of above the threshold
+        if len(filtered_vif_scores) == 0:
+            break
+
+        # exclude the metric with the highest VIF score
+        metric_to_exclude = list(filtered_vif_scores['Feature'].head(1))[0]
+
+        print('> Step', count, '- exclude', str(metric_to_exclude))
+        count = count + 1
+
+        selected_features = list(set(selected_features) - set([metric_to_exclude]))
+
+        X_AS_train = X_AS_train.loc[:, selected_features]
+    print('Finally, according to Part 2 of AutoSpearman,', AS_metrics, 'are selected.')
+    return X_AS_train
 
 
 def get_base_prefix_compat():
@@ -56,48 +164,105 @@ def data_validation(data):
     return valid
 
 
-def get_default_data_and_model():
+def filter_rules(rules, X_explain):
+    """Get rules that are actually applied to the commit
+
+    Parameters
+    ----------
+    rules : :obj:`pandas.core.frame.DataFrame`
+        Rules data under the column called 'rule' inside Rules DF generated by RuleFit
+    X_explain : :obj:`pandas.core.frame.DataFrame`
+        Features to be explained by the local RuleFit model, can be seen as X_test
+
+    Returns
+    -------
+    :obj:`pandas.core.frame.DataFrame`
+        A DataFrame that contains filtered rules
+    """
+
+    def eval_rule(rule, x_df):
+        """Parsing pattern of one rule (a single row under the rule column inside Rules DF generated by RuleFit)
+
+        Note. Only support names of rules (feature names) that has CamelCase pattern with or without underscore
+        e.g. 'ThisIsLegal_Name', 'this_is_legal_name', 'thisislegalname', 'ThisIsLegalName'
+        Name like 'this is not a legal name', consider using '_' instead of ' ' to concatenate
+        """
+        var_in_rule = list(set(re.findall('[a-z_*A-Z]+', rule)))
+        rule = re.sub(r'\b=\b', '==', rule)
+        if 'or' in var_in_rule:
+            var_in_rule.remove('or')
+        rule = rule.replace('&', 'and')
+        eval_result_list = []
+        for i in range(0, len(x_df)):
+            x = x_df.iloc[[i]]
+            var_dict = {}
+            for var in var_in_rule:
+                var_dict[var] = float(x[var])
+            eval_result = eval(rule, var_dict)
+            eval_result_list.append(eval_result)
+        return eval_result_list
+
+    # select rules that (1) have positive coefficient values and (2) have positive importance scores
+    rules = rules[(rules['type'] == 'rule') & (rules['coef'] > 0) & (rules['importance'] > 0)]
+    rules_list = list(rules['rule'])
+    rule_eval_result = []
+    # for each rule, check whether such rules apply to the actual instance to be explained
+    # Note.
+    # you may pass rules variable to eval_rule() to get the result of all rules
+    # never try, but if it works please tell me.
+    for r in rules_list:
+        # X_explain must be a dataframe of 1 row
+        py_exp_pred = eval_rule(r, X_explain)[0]
+        rule_eval_result.append(py_exp_pred)
+    new_col = {'is_satisfy_instance': rule_eval_result}
+    df = pd.DataFrame(data=new_col)
+    rules = pd.concat([rules, df], axis=1)
+    # select rules that apply to the actual instance to be explained
+    # Note. can't use 'is' as comparing operator because of pandas
+    rules = rules.loc[rules['is_satisfy_instance'] == True]
+    # sort rules according to their importance scores
+    sorted_rules = rules.sort_values(by='importance', ascending=False)
+    return sorted_rules
+
+
+def get_dflt():
+    """Obtain the default data and model
+
+    Returns
+    -------
+    :obj:`dict`
+        A dictionary wrapping all default data and model
+    """
     this_dir, _ = os.path.split(__file__)
-    path_train = this_dir + "/default_data/activemq-5.0.0.zip"
-    path_test = this_dir + "/default_data/activemq-5.1.0.zip"
+    path_dflt = this_dir + "/default_data/dflt_data.pkl"
+    path_rf_model = this_dir + "/rf_models/jit_rf_model.pkl"
+    path_explain = this_dir + "/default_data/explained_instance.pkl"
     if INSIDE_VIRTUAL_ENV:
         cwd = os.getcwd()
-        path_train = cwd + "/tests/pyexplainer_test_data/activemq-5.0.0.zip"
-        path_test = cwd + "/tests/pyexplainer_test_data/activemq-5.1.0.zip"
-    training_data = pd.read_csv(path_train, index_col='File')
-    dependent_vars = training_data.columns[-4]
-    selected_features = ["ADEV", "AvgCyclomaticModified", "AvgEssential", "AvgLineBlank", "AvgLineComment",
-                         "CountClassBase", "CountClassCoupled", "CountClassDerived", "CountDeclClass",
-                         "CountDeclClassMethod", "CountDeclClassVariable", "CountDeclInstanceVariable",
-                         "CountDeclMethodDefault", "CountDeclMethodPrivate", "CountDeclMethodProtected",
-                         "CountDeclMethodPublic", "CountInput_Mean", "CountInput_Min", "CountOutput_Min", "MAJOR_LINE",
-                         "MaxInheritanceTree", "MaxNesting_Min", "MINOR_COMMIT", "OWN_COMMIT", "OWN_LINE",
-                         "PercentLackOfCohesion", "RatioCommentToCode"]
-    all_cols = training_data.columns
-    for col in all_cols:
-        if col not in selected_features:
-            all_cols = all_cols.drop(col)
-    independent_vars = all_cols
-    X_train = training_data.loc[:, independent_vars]
-    y_train = training_data.loc[:, dependent_vars]
+        path_dflt = cwd + "/tests/default_data/dflt_data.pkl"
+        path_rf_model = cwd + "/tests/rf_models/jit_rf_model.pkl"
+        path_explain = cwd + "/tests/default_data/explained_instance.pkl"
+    with open(path_dflt, 'rb') as f:
+        dflt_data = pickle.load(f)
+    with open(path_rf_model, 'rb') as f:
+        rf_model = pickle.load(f)
+    with open(path_explain, 'rb') as f:
+        explained_instance = pickle.load(f)
 
-    blackbox_model = RandomForestClassifier(max_depth=3, random_state=0)
-    blackbox_model.fit(X_train, y_train)
-
-    testing_data = pd.read_csv(path_test, index_col='File')
-    X_test = testing_data.loc[:, independent_vars]
-    y_test = testing_data.loc[:, dependent_vars]
-
-    sample_explain_index = 0
-    X_explain = X_test.iloc[[sample_explain_index]]
-    y_explain = y_test.iloc[[sample_explain_index]]
-    return {'X_train': X_train,
-            'y_train': y_train,
-            'indep': independent_vars,
-            'dep': dependent_vars,
-            'blackbox_model': blackbox_model,
-            'X_explain': X_explain,
-            'y_explain': y_explain}
+    full_ft_names = ['Number Of Lines Added', 'Number Of Modified Directories',
+                     'Number Of Modified Subsystems', 'Entropy',
+                     'Number Of Review Revisions', 'Review Timespan',
+                     'Number of Self Approvals', 'Number of Past Developers',
+                     'Time Since Last Modification', 'Number of Reviewrs',
+                     'Relative Reviewer Experience', 'Author Awareness', 'Reviewer Awareness']
+    return {'X_train': dflt_data['x_train'],
+            'y_train': dflt_data['y_train'],
+            'indep': dflt_data['indep'],
+            'dep': dflt_data['dep'],
+            'blackbox_model': rf_model,
+            'X_explain': explained_instance['X_explain'],
+            'y_explain': explained_instance['y_explain'],
+            'full_ft_names': full_ft_names}
 
 
 def id_generator(size=15, random_state=check_random_state(None)):
@@ -165,6 +330,8 @@ class PyExplainer:
         Classification labels, default = ['Clean', 'Defect']
     top_k_rules : :obj:`int`
         Number of top positive and negative rules to be retrieved
+    full_ft_names : :obj:`list`
+        A list containing full feature names inside X_train
     """
 
     def __init__(self,
@@ -174,7 +341,8 @@ class PyExplainer:
                  dep,
                  blackbox_model,
                  class_label=['Clean', 'Defect'],
-                 top_k_rules=3):
+                 top_k_rules=3,
+                 full_ft_names=[]):
         if isinstance(X_train, pd.core.frame.DataFrame):
             self.X_train = X_train
         else:
@@ -220,6 +388,15 @@ class PyExplainer:
         else:
             print("top_k_rules should be type 'int'")
             raise TypeError
+        if full_ft_names:
+            short_ft_names = X_train.columns.to_list()
+            # length of short ft names and full ft names should be the same
+            if len(short_ft_names) != len(full_ft_names):
+                print("list of short feature names and list of full feature names should have the same length!")
+                raise ValueError
+            self.full_ft_names = dict(zip(short_ft_names, full_ft_names))
+        else:
+            self.full_ft_names = {}
 
         self.bullet_data = [{}]
         self.risk_data = [{}]
@@ -228,7 +405,11 @@ class PyExplainer:
         self.X_explain = None
         self.y_explain = None
 
-    def auto_spearman(self, apply_to_X_train=True, correlation_threshold=0.7, correlation_method='spearman', VIF_threshold=5):
+    def auto_spearman(self,
+                      apply_to_X_train=True,
+                      correlation_threshold=0.7,
+                      correlation_method='spearman',
+                      VIF_threshold=5):
         """An automated feature selection approach that address collinearity and multicollinearity.
         For more information, please kindly refer to the `paper <https://ieeexplore.ieee.org/document/8530020>`_.
 
@@ -243,99 +424,19 @@ class PyExplainer:
         VIF_threshold : :obj:`int`
             Threshold value of VIF score.
         """
-        X_AS_train = self.X_train.copy()
-        AS_metrics = X_AS_train.columns
-        count = 1
-
-        # (Part 1) Automatically select non-correlated metrics based on a Spearman rank correlation test.
-        print('(Part 1) Automatically select non-correlated metrics based on a Spearman rank correlation test')
-        while True:
-            corrmat = X_AS_train.corr(method=correlation_method)
-            top_corr_features = corrmat.index
-            abs_corrmat = abs(corrmat)
-
-            # identify correlated metrics with the correlation threshold of the threshold
-            highly_correlated_metrics = ((corrmat > correlation_threshold) | (corrmat < -correlation_threshold)) & (
-                        corrmat != 1)
-            n_correlated_metrics = np.sum(np.sum(highly_correlated_metrics))
-            if n_correlated_metrics > 0:
-                # find the strongest pair-wise correlation
-                find_top_corr = pd.melt(abs_corrmat, ignore_index=False)
-                find_top_corr.reset_index(inplace=True)
-                find_top_corr = find_top_corr[find_top_corr['value'] != 1]
-                top_corr_index = find_top_corr['value'].idxmax()
-                top_corr_i = find_top_corr.loc[top_corr_index, :]
-
-                # get the 2 correlated metrics with the strongest correlation
-                correlated_metric_1 = top_corr_i[0]
-                correlated_metric_2 = top_corr_i[1]
-                print('> Step', count, 'comparing between', correlated_metric_1, 'and', correlated_metric_2)
-
-                # compute their correlation with other metrics outside of the pair
-                correlation_with_other_metrics_1 = np.mean(abs_corrmat[correlated_metric_1][
-                                                               [i for i in top_corr_features if
-                                                                i not in [correlated_metric_1, correlated_metric_2]]])
-                correlation_with_other_metrics_2 = np.mean(abs_corrmat[correlated_metric_2][
-                                                               [i for i in top_corr_features if
-                                                                i not in [correlated_metric_1, correlated_metric_2]]])
-                print('>>', correlated_metric_1, 'has the average correlation of',
-                      np.round(correlation_with_other_metrics_1, 3), 'with other metrics')
-                print('>>', correlated_metric_2, 'has the average correlation of',
-                      np.round(correlation_with_other_metrics_2, 3), 'with other metrics')
-                # select the metric that shares the least correlation outside of the pair and exclude the other
-                if correlation_with_other_metrics_1 < correlation_with_other_metrics_2:
-                    exclude_metric = correlated_metric_2
-                else:
-                    exclude_metric = correlated_metric_1
-                print('>>', 'Exclude', exclude_metric)
-                count = count + 1
-                AS_metrics = list(set(AS_metrics) - set([exclude_metric]))
-                X_AS_train = X_AS_train[AS_metrics]
-            else:
-                break
-
-        print('According to Part 1 of AutoSpearman,', AS_metrics, 'are selected.')
-
-        # (Part 2) Automatically select non-correlated metrics based on a Variance Inflation Factor analysis.
-        print('(Part 2) Automatically select non-correlated metrics based on a Variance Inflation Factor analysis')
-
-        # Prepare a dataframe for VIF
-        X_AS_train = add_constant(X_AS_train)
-
-        selected_features = X_AS_train.columns
-        count = 1
-        while True:
-            # Calculate VIF scores
-            vif_scores = pd.DataFrame([variance_inflation_factor(X_AS_train.values, i)
-                                       for i in range(X_AS_train.shape[1])],
-                                      index=X_AS_train.columns)
-            # Prepare a final dataframe of VIF scores
-            vif_scores.reset_index(inplace=True)
-            vif_scores.columns = ['Feature', 'VIFscore']
-            vif_scores = vif_scores.loc[vif_scores['Feature'] != 'const', :]
-            vif_scores.sort_values(by=['VIFscore'], ascending=False, inplace=True)
-
-            # Find features that have their VIF scores of above the threshold
-            filtered_vif_scores = vif_scores[vif_scores['VIFscore'] >= VIF_threshold]
-
-            # Terminate when there is no features with the VIF scores of above the threshold
-            if len(filtered_vif_scores) == 0:
-                break
-
-            # exclude the metric with the highest VIF score
-            metric_to_exclude = list(filtered_vif_scores['Feature'].head(1))[0]
-
-            print('> Step', count, '- exclude', str(metric_to_exclude))
-            count = count + 1
-
-            selected_features = list(set(selected_features) - set([metric_to_exclude]))
-
-            X_AS_train = X_AS_train.loc[:, selected_features]
-
-        print('Finally, according to Part 2 of AutoSpearman,', AS_metrics, 'are selected.')
+        X_AS_train = AutoSpearman(correlation_threshold, correlation_method, VIF_threshold)
         if apply_to_X_train:
             self.set_X_train(X_AS_train)
-            print('X_train data inside PyExplainer was updated based on the selected features above')
+            # if there is data of full feature names
+            if self.get_full_ft_names():
+                full_ft_names = self.get_full_ft_names()
+                new_full_ft_names = {}
+                for key in X_AS_train.columns.to_list():
+                    new_full_ft_names[key] = full_ft_names[key]
+                self.set_full_ft_names(new_full_ft_names)
+            return print('X_train data inside PyExplainer was updated based on the selected features above')
+        else:
+            return X_AS_train
 
     def explain(self,
                 X_explain,
@@ -453,10 +554,16 @@ class PyExplainer:
         rules = local_rulefit_model.get_rules()
         rules = rules[rules.coef != 0].sort_values("importance", ascending=False)
         rules = rules[rules.type == 'rule'].sort_values("importance", ascending=False)
+        positive_filtered_rules = filter_rules(rules, X_explain)
 
-        top_k_positive_rules = rules[rules.coef > 0].sort_values("importance", ascending=False).head(top_k)
+        # positive rules
+        top_k_positive_rules = positive_filtered_rules.loc[positive_filtered_rules['coef'] > 0].sort_values(
+            "importance", ascending=False).head(top_k)
         top_k_positive_rules['Class'] = self.class_label[1]
-        top_k_negative_rules = rules[rules.coef < 0].sort_values("importance", ascending=False).head(top_k)
+        top_k_positive_rules = positive_filtered_rules.reset_index()
+
+        # negative rules
+        top_k_negative_rules = rules.loc[rules['coef'] < 0].sort_values("importance", ascending=False).head(top_k)
         top_k_negative_rules['Class'] = self.class_label[0]
 
         rule_obj = {'synthetic_data': synthetic_instances,
@@ -489,56 +596,63 @@ class PyExplainer:
         bullet_data = []
 
         for i in range(len(parsed_rule_object['top_tofollow_rules'])):
+            # sample data of tmp_rule
             # {'variable': 'MAJOR_COMMIT', 'lessthan': True, 'value': '1.550000011920929'}
             tmp_rule = parsed_rule_object['top_tofollow_rules'][i]
-            tmp_min = int((min_max_values['min_values'][tmp_rule['variable']]))
-            tmp_max = int(round(min_max_values['max_values'][tmp_rule['variable']]))
-            tmp_interval = (tmp_max - tmp_min) / 10.0
-            tmp_threshold_value = round(float(tmp_rule['value']), 2)
             tmp_actual_value = round(X_explain[tmp_rule['variable']][0], 2)
-            tmp_markers = [tmp_actual_value]
-            plot_min = int(round(max(tmp_min, tmp_threshold_value - tmp_interval), 0))  # * 1.0
-            plot_max = int(round(min(tmp_max, tmp_threshold_value + tmp_interval), 0))  # * 1.0
+            separation_point = float(tmp_rule['value'])
 
-            # keep marker in the range
-            if tmp_markers[0] > plot_max:
-                plot_max = tmp_markers[0]
-            elif tmp_markers[0] < plot_min:
-                plot_min = tmp_markers[0]
+            if tmp_actual_value < 0 or separation_point < 0:
+                print("""actual value of %s < 0, currently do not support this type of rule""" % tmp_rule['variable'])
+
+            tmp_markers = [tmp_actual_value]
+
+            # currently, always place marker in the middle of bullet bars
+            plot_min = 0
+            diff_actual_min = abs(tmp_actual_value - plot_min)
+            plot_max = tmp_actual_value + diff_actual_min
+            if separation_point > plot_max:
+                plot_max += separation_point - plot_max
+                plot_max += diff_actual_min
 
             diff_plot_max_min = plot_max - plot_min
             tmp_subtitle_text = 'Actual = ' + str(tmp_actual_value)
             tmp_ticks = [plot_min, plot_max]
 
-            if plot_max - plot_min <= 10:
+            if diff_plot_max_min <= 1:
+                tmp_step = [0.01]
+            elif diff_plot_max_min >= 1000:
+                tmp_step = [10]
+            elif diff_plot_max_min <= 3:
                 tmp_step = [0.1]
             else:
                 tmp_step = [1]
 
             bullet_total_width = 450
+            tmp_start_points = [0, round((separation_point / diff_plot_max_min if diff_plot_max_min else 0)
+                                         * bullet_total_width, 4)]
+            tmp_widths = [round(tmp_start_points[1] - plot_min, 4), round(bullet_total_width - tmp_start_points[1], 4)]
 
-            tmp_start_points = [0, round((tmp_threshold_value - plot_min) / diff_plot_max_min * bullet_total_width
-                                         if diff_plot_max_min * bullet_total_width else 0, 0)]
-
-            tmp_widths = [round((tmp_threshold_value - plot_min) / diff_plot_max_min * bullet_total_width
-                                if diff_plot_max_min * bullet_total_width else 0, 0),
-                          round((plot_max - tmp_threshold_value) / diff_plot_max_min * bullet_total_width
-                                if diff_plot_max_min * bullet_total_width else 0, 0)]
-
-            id = '#' + str(i + 1)
+            title_id = '#' + str(i + 1)
             var_name = str(tmp_rule['variable'])
+            var_ref = var_name
+            # todo - use get
+            # check if there is mapping for full feature names
+            if self.full_ft_names:
+                # todo - use get
+                full_ft_names = self.full_ft_names
+                var_name = full_ft_names[var_name]
             if tmp_rule['lessthan']:
-                # lessthan == TRUE:
                 # The rule suggest to decrease the values to less than a certain threshold
-                tmp_title_text = id + ' Decrease the values of ' + \
-                                 var_name + ' to less than ' + \
+                tmp_title_text = title_id + ' The value of ' + \
+                                 var_name + ' is more than ' + \
                                  str(tmp_actual_value)
                 tmp_colors = ["#a6d96a", "#d7191c"]
             else:
                 # lessthan == FALSE:
                 # The rule suggest to increase the values to more than a certain threshold
-                tmp_title_text = id + ' Increase the values of ' + \
-                                 var_name + ' to more than ' + \
+                tmp_title_text = title_id + ' The value of ' + \
+                                 var_name + ' is less than ' + \
                                  str(tmp_actual_value)
                 tmp_colors = ["#d7191c", "#a6d96a"]
 
@@ -551,7 +665,7 @@ class PyExplainer:
                 "widths": tmp_widths,
                 "colors": tmp_colors,
                 "markers": tmp_markers,
-                "varRef": var_name,
+                "varRef": var_ref,
             })
         return bullet_data
 
@@ -580,7 +694,7 @@ class PyExplainer:
         <script>%s</script>
         """ % (d3_js, bullet_js)
 
-        main_title = "What to do to decrease the risk of having defects?"
+        main_title = "Why this commit is predicted as defect-introducing?"
         title = """
         <div style="position: relative; top: 0; width: 100vw; left: 20vw;">
             <b>%s</b>
@@ -589,7 +703,6 @@ class PyExplainer:
 
         unique_id = id_generator()
         bullet_data = to_js_data(self.__get_bullet_data())
-        risk_data = to_js_data(self.__get_risk_data())
 
         d3_operation_script = """
         <script>
@@ -602,30 +715,6 @@ class PyExplainer:
 
         var bulletData = %s
 
-        var riskData = %s
-
-        // define the color of the box
-        var boxColor = "box green";
-        var riskPred = riskData[0].riskPred[0];
-        if (riskPred.localeCompare("Defect")==0) {
-            boxColor = "box orange";
-        }
-
-        // append risk prediction and risk score
-        d3.select("#d3-target-bullet-%s")
-          .append("div")
-          .attr("class", "riskPred")
-          .data(riskData)
-          .text((d) => d.riskPred)
-          .append("div")
-          .attr("class", boxColor);
-
-        d3.select("#d3-target-bullet-%s")
-          .append("div")
-          .attr("class", "riskScore")
-          .data(riskData)
-          .text((d) => "Risk Score: " + d.riskScore);
-
         var svg = d3
           .select("#d3-target-bullet-%s")
           .selectAll("svg")
@@ -659,7 +748,7 @@ class PyExplainer:
           .text((d) => d.subtitle);
 
         </script>
-        """ % (bullet_data, risk_data, unique_id, unique_id, unique_id)
+        """ % (bullet_data, unique_id)
 
         html = """
         <!DOCTYPE html>
@@ -746,7 +835,7 @@ class PyExplainer:
             dist = np.linalg.norm(
                 trainset_normalize.sub(np.array(case)), axis=1)
             # Convert distance to a similarity score
-            similarity = np.exp( -(dist ** 2) / (2 * (width ** 2)) )
+            similarity = np.exp(-(dist ** 2) / (2 * (width ** 2)))
             dist_df['dist'] = similarity
             dist_df['t_target'] = target_train
             # get the unique classes of the training set
@@ -1014,6 +1103,16 @@ class PyExplainer:
                  "riskPred": [self.class_label[risk_pred]]
                  }]
 
+    def get_full_ft_names(self):
+        """getter of self.full_ft_names
+
+        Returns
+        ----------
+        :obj:`list`
+            A list of full feature names in X_train following the same order as X_train
+        """
+        return self.full_ft_names
+
     def get_risk_pred(self):
         """Retrieve the risk prediction from risk_data
 
@@ -1276,6 +1375,16 @@ class PyExplainer:
         # update the right text
         self.update_right_text(right_text)
 
+    def set_full_ft_names(self, full_ft_names):
+        """Setter of full_ft_names
+
+        Parameters
+        ----------
+        full_ft_names : :obj:`list`
+            A list of full feature names in X_train following the same order as X_train
+        """
+        self.full_ft_names = full_ft_names
+
     def set_top_k_rules(self, top_k_rules):
         """Setter of top_k_rules
 
@@ -1314,12 +1423,6 @@ class PyExplainer:
         display(widgets.HBox(items))
         self.run_bar_animation()
 
-        # display sliders
-        sliders = self.generate_sliders()
-        for slider in sliders:
-            slider.observe(self.on_value_change, names='value')
-            display(slider)
-
         bullet_out = self.bullet_output
         bullet_out.clear_output()
         display(bullet_out)
@@ -1328,6 +1431,12 @@ class PyExplainer:
             html = self.generate_html()
             display(HTML(html))
 
+        # display sliders
+        sliders = self.generate_sliders()
+        for slider in sliders:
+            slider.observe(self.on_value_change, names='value')
+            display(slider)
+
     def update_risk_score(self, risk_score):
         """Update the risk score value inside the risk_data
 
